﻿Imports MediaCy.Addins.DataCollector
Imports MediaCy.Addins.Measurements
Imports MediaCy.IQL.Features
Imports MediaCy.IQL.ObjectManager
Imports MediaCy.IQL.Features.mcMeasurementDependencyFlags
Imports MediaCy.IQL.Features.mcMeasurementParent
Imports MediaCy.IQL.Features.mcMeasurementCategoryFlags
Imports MediaCy.IQL.Features.mcMeasurementAttributeID
Imports MediaCy.IQL.Features.mcMeasurementResultType
Imports System.Collections

Public Enum eEdgeType
    Steps=0
    Dark=1
    Bright=2
End Enum

Public Enum eImageType
    Native=0
    Binary=1
End Enum

Public Enum eMeasureType
    OnePerSegment=0
    Equidistant=1
End Enum


Public Class PresetROI
    Public Name As String
    Public Rect As SINGLERECT
    Public Sub New()
        Name=""
    End Sub

    Public Sub SetStr(S As String())
        Name=S(0)
        Rect.left=Single.Parse(S(1))
        Rect.top=Single.Parse(S(2))
        Rect.right=Single.Parse(S(3))
        Rect.bottom=Single.Parse(S(4))
    End Sub
    Public Sub SetRect(S As String,r As SINGLERECT)
        Name=S
        Rect=r
    End Sub
    Public Function ToString() As String
        Return String.Format("{0},{1},{2},{3},{4}",Name,Rect.left,Rect.top,Rect.right,Rect.bottom)
    End Function
End Class

Public Class PresetROIList
    'Inherits System.Collections.Generic.List(Of PresetROI)
    Public Property ROIs As System.Collections.Generic.List(Of PresetROI)
    Public Sub New()
        ROIs =New System.Collections.Generic.List(Of PresetROI)
    End Sub

End Class

Public Class Options
    Public MaxFiberWidth As Double
    Public MinFiberLength As Double
    Public IncrStep As Double
    Public SmoothingSize As Double
    Public MinArea As Double
    Public Interactive As Boolean
    Public EdgeType As eEdgeType
    Public ImageType As eImageType
    Public MeasureType As eMeasureType
    Public MaskImage As McImage
    Public CalibrationName As String
    Public PresetROIs As String
    Public ActivePresetROI As Integer
    Public ROIsList As New PresetROIList
    Public CollectToDC As Boolean
    Public SourceFolder As String
    Public OutputFolder As String
    Public SaveSnapshot As Boolean
    Public SaveData As Boolean
    Public KeepResultImage As Boolean
    Public UseCurrentOptions As Boolean
    Public MonoInterpretation As Boolean
    Public DarkFibers As Boolean
    Public UseCircles As Boolean'use circles else lines
    Public CorrectBackground As Boolean
    Public FillFibers As Boolean'apply close/open filter to fill hollow fibers
    Public Threshold As Double
    Public CentralLine As Boolean

    Public Sub New()
        Reset()
        Load()
    End Sub

    Public Sub Reset()
        MaxFiberWidth =50
        MinFiberLength=30
        IncrStep=20
        SmoothingSize = 5
        MinArea =0
        EdgeType=eEdgeType.Steps
        ImageType=eImageType.Native
        MeasureType=eMeasureType.OnePerSegment
        Interactive=False
        CalibrationName = ""
        PresetROIs=""
        ActivePresetROI=0
        CollectToDC=True
        SourceFolder=""
        OutputFolder="Data"
        SaveSnapshot=False
        SaveData=False
        KeepResultImage=False
        UseCurrentOptions=False
        MonoInterpretation=True
        DarkFibers=True
        UseCircles=False
        CorrectBackground=True
        FillFibers=False
        Threshold=2
        CentralLine=False
    End Sub

    Private Function MySettings() As McSettings
        Return ThisApplication.Settings("FiberThicknessApp")
    End Function

    Public Sub Load(Optional st As McSettings=Nothing)
        'init options
        If st Is Nothing Then
            st=MySettings()
        End If
        Reset()
        With st
            MaxFiberWidth =.Get("MaxFiberWidth", MaxFiberWidth)
            MinFiberLength=.Get("MinFiberLength",MinFiberLength)
            IncrStep=.Get("IncrStep",IncrStep)
            SmoothingSize = .Get("SmoothingSize", SmoothingSize)
            MinArea =.Get("MinArea", MinArea)
            Interactive=.Get("Interactive",Interactive)
            EdgeType=.Get("EdgeType",CInt(EdgeType))
            ImageType=.Get("ImageType",CInt(ImageType))
            MeasureType=.Get("MeasureType",CInt(MeasureType))
            CalibrationName = .Get("CalibrationName", "")
            PresetROIs=.Get("PresetROIs","")
            ROIsToList
            ActivePresetROI=.Get("ActivePresetROI",0)
            CollectToDC=.Get("CollectToDC",True)
            SourceFolder=.Get("SourceFolder","")
            OutputFolder=.Get("OutputFolder","Data")
            SaveSnapshot=.Get("SaveSnapshot",False)
            SaveData=.Get("SaveData",False)
            KeepResultImage=.Get("KeepResultImage",False)
            UseCurrentOptions=.Get("UseCurrentOptions",False)
            MonoInterpretation=.Get("MonoInterpretation",True)
            DarkFibers=.Get("DarkFibers",DarkFibers)
            UseCircles=.Get("UseCircles",UseCircles)
            CorrectBackground=.Get("CorrectBackground",CorrectBackground)
            FillFibers=.Get("FillFibers",FillFibers)
            Threshold=.Get("Threshold",Threshold)
            CentralLine=.Get("CentralLine",CentralLine)
        End With
    End Sub

    Public Sub ROIsToList
        ROIsList=New PresetROIList
        Dim ss As String()=PresetROIs.Split(";"c)
        For Each s As String In ss
            Dim s2 As String()=s.Split(","c)
            If s2.GetUpperBound(0)=4 Then
                Dim pr As New PresetROI
                pr.SetStr(s2)
                ROIsList.ROIs.Add(pr)
            End If
        Next
    End Sub

    Public Sub ROIsToString
        PresetROIs = ""
        For Each r As PresetROI In ROIsList.ROIs
            If PresetROIs<>"" Then PresetROIs+=";"
            PresetROIs+=r.ToString
        Next
    End Sub

    Public Sub Save(Optional st As McSettings=Nothing)
        'save options
        If st Is Nothing Then
            st=MySettings()
        End If
        With st
            .Set("MaxFiberWidth", MaxFiberWidth)
            .Set("MinFiberLength",MinFiberLength)
            .Set("IncrStep",IncrStep)
            .Set("SmoothingSize", SmoothingSize)
            .Set("MinArea", MinArea)
            .Set("Interactive",Interactive)
            .Set("EdgeType",CInt(EdgeType))
            .Set("ImageType",CInt(ImageType))
            .Set("MeasureType",CInt(MeasureType))
            .Set("CalibrationName", CalibrationName)
            ROIsToString
            .Set("PresetROIs",PresetROIs)
            .Set("ActivePresetROI",ActivePresetROI)
            .Set("CollectToDC",CollectToDC)
            .Set("SourceFolder",SourceFolder)
            .Set("OutputFolder",OutputFolder)
            .Set("SaveSnapshot",SaveSnapshot)
            .Set("SaveData",SaveData)
            .Set("KeepResultImage",KeepResultImage)
            .Set("UseCurrentOptions",UseCurrentOptions)
            .Set("MonoInterpretation",MonoInterpretation)
            .Set("DarkFibers",DarkFibers)
            .Set("CorrectBackground",CorrectBackground)
            .Set("FillFibers",FillFibers)
            .Set("Threshold",Threshold)
            .Set("CentralLine",CentralLine)
        End With
    End Sub

    Public Sub SaveToFile(ByVal FileName As String)
        Dim MySettings As MediaCy.IQL.Application.McSettings = New MediaCy.IQL.Application.McSettings(ThisApplication)
        Save(MySettings)
        MySettings.Save(FileName)
    End Sub

    Public Sub LoadFromFile(ByVal FileName As String)
        Dim MySettings As MediaCy.IQL.Application.McSettings = New MediaCy.IQL.Application.McSettings(ThisApplication)
        MySettings.Load(FileName)
        Load(MySettings)
    End Sub

End Class

Public Module AppModule
    Public opt As New Options

    Public Function SaveOptions()
        Dim d As New System.Windows.Forms.SaveFileDialog()
        With d
            .InitialDirectory = ThisApplication.Path(mcPathType.mcptConfigurationFiles)
            .Filter = "Fiber Thickness Options (*.fto)|*.fto"
            .FilterIndex = 1
            .RestoreDirectory = False
            .CheckPathExists = False
            If .ShowDialog() = System.Windows.Forms.DialogResult.OK Then
                If .FileName.Length > 0 Then
                    opt.SaveToFile(.FileName)
                End If
            End If
        End With
    End Function

    Public Function LoadOptions() As Boolean
        Dim d As New System.Windows.Forms.OpenFileDialog()
        With d
            .InitialDirectory = ThisApplication.Path(mcPathType.mcptConfigurationFiles)
            .Filter = "Fiber Thickness Options (*.fto)|*.fto"
            .FilterIndex = 1
            .RestoreDirectory = False
            .CheckPathExists = True
            .CheckFileExists = True
            If .ShowDialog() = System.Windows.Forms.DialogResult.OK Then
                If .FileName.Length > 0 Then
                    opt.LoadFromFile(.FileName)
                    Return True
                End If
            End If
        End With
        Return False
    End Function

    Public Sub MeasureFibersOnActiveImage()
        'save settings
        opt.Save()
        Try
            MeasureFiberThickness()
        Catch Ex As System.Exception
            MsgBox(Ex.ToString)
        Finally
        End Try
    End Sub

    Private Function GetMcImage(src As Object) As McImage
        If TypeOf src Is McImage Then
            Return src
        ElseIf TypeOf src Is MediaCy.IQL.Sets.McImageSet Then
            Dim imSet As MediaCy.IQL.Sets.McImageSet = src
            Return imSet.GetSourceAtLocation(imSet.GetCurrentLocation(), MediaCy.IQL.Sets.mcImageSetLocationMatchTypes.mcislmtNearest).Image
        End If
        Return Nothing
    End Function

    Public Function BinarizeImage(im As McImage) As McImage
        If im.NumberOfChannels > 1 Then
            If True Then'opt.MonoInterpretation Then
                'convert to gray
                Dim newType As MediaCy.IQL.Engine.McImageType = im.Type.DuplicateGivenType
                newType.Initialize(newType.Organization, newType.PixelDataType, mcInterpretation.mciMonochrome, newType.BitsPerChannel, 1)
                Dim imGray As McImage = im.Convert(newType, mcImageCreateFlags.mcicfNotVisible Or mcImageCreateFlags.mcicfNoAddToCollection)
                TmpImages.Add(imGray)
                'im.Close
                im = imGray
            End If
        End If

        'processing
        Dim Flags As mcImageCreateFlags = GetCreateFlags

        Dim bDarkFibers As Boolean=opt.DarkFibers
        Dim fSize As Integer=opt.MaxFiberWidth
        Dim lowPassSize As Integer=opt.SmoothingSize

        'extract backround
                'Dim dupl As McImage=im.Morpho.ExtractBackground(fSize,Not bDarkFibers,Flags)
        Dim dupl As McImage=im.Duplicate(Flags)
        TmpImages.Add(dupl)
        If opt.FillFibers Then
            Dim ftSize As Integer=fSize*0.666
            If bDarkFibers Then
                'apply Open filter
                im.Filter3D.Open3D(ftSize,ftSize,1,1)
            Else
                'apply Close filter
                im.Filter3D.Close3D(ftSize,ftSize,1,1)
            End If
        End If

        If opt.CorrectBackground Then
            dupl.LargeFilter.LowPassLarge(fSize*2,fSize*2,3)

            If bDarkFibers Then
                dupl.Op.Sub(im)
                'im.Close
                im=dupl
            Else
                im.Op.Sub(dupl)
                'dupl.Close
            End If
        End If

        'im.LargeFilter.BandPassLarge(lowPassSize,1000,3,8)
        If lowPassSize>1 Then
            im.LargeFilter.LowPassLarge(lowPassSize,lowPassSize,3)
        End If
        Dim lgSize As Integer=1000
        If opt.CorrectBackground Then
            lgSize = 1000
            im.LargeFilter.EdgePlusLarge(lgSize,lgSize,1,2)
        Else
            lgSize = fSize*2+1
            If opt.DarkFibers Then
                im.LargeFilter.EdgeMinusLarge(lgSize,lgSize,1,2)
            Else
                im.LargeFilter.EdgePlusLarge(lgSize,lgSize,1,2)
            End If
        End If

        Dim thr As Double=opt.Threshold
        Dim mult As Double=10000
        im.Op.Sub(thr)
        im.Op.Mult(mult)'binarize at 1

        Return im
    End Function

    'keep list of temp images untill the end, to avoid garbage colleciton in the middle
    Private TmpImages As New System.Collections.ArrayList

    Public Sub ClearTempImages()
        'MsgBox("Before cleanup")
        System.GC.Collect
        System.GC.WaitForPendingFinalizers
        'MsgBox("After first GC")
        For Each im As Object In TmpImages
            If TypeOf(im) Is McImage Then
                im.MeasurementsData.Reset
                im.Modified=False
                im.Close
            Else
            End If
        Next
        'MsgBox("After images close")
        TmpImages.Clear
        System.GC.Collect
        System.GC.WaitForPendingFinalizers
        'MsgBox("After second GC")
    End Sub


    Public Sub MeasureFiberThickness(Optional bFolder As Boolean = False, Optional bPreviewBin As Boolean=False)
        If bPreviewBin AndAlso opt.ImageType=eImageType.Binary Then
            Exit Sub'don't do preview on bin
        End If
        ClearTempImages()
        Dim Fibers As New SimpleScript
        Dim doc0, image1, doc2, window1
        Dim image0 As McImage
        Dim binImage As McImage

        With Application.DocumentCommands.Active(Fibers)
            .Run(doc0)
        End With

        image0 = MediaCy.Commands.Image.GetImage(doc0)
        If image0 Is Nothing Then Exit Sub

        'reset measurements
        image0.MeasurementsData.Reset

        'apply calibration
        If opt.CalibrationName.Length > 0 Then
            'try to find this calibration
            Dim selCal As MediaCy.IQL.Calibrations.McSpatialCalib = Nothing
            Dim cals As MediaCy.IQL.Calibrations.McSpatialCalibs = ThisApplication.SpatialCalibs
            For Each cal As MediaCy.IQL.Calibrations.McSpatialCalib In cals
                If cal.Name = opt.CalibrationName Then
                    selCal = cal
                    Exit For
                End If
            Next
            If selCal IsNot Nothing Then
                image0.SpatialCalibration = selCal
            End If
        End If

        Dim pr As PresetROI=Nothing
        Dim roi As McRegions
        Dim rl As Integer=0,rt As Integer=0,rr As Integer,rb As Integer
        'apply ROI
        If opt.ActivePresetROI > 0 AndAlso ((opt.ActivePresetROI - 1) < opt.ROIsList.ROIs.Count) Then
            pr = opt.ROIsList.ROIs(opt.ActivePresetROI - 1)
            roi = image0.Aoi
            roi.Reset

            roi.SetBox(0, pr.Rect.left, pr.Rect.top, pr.Rect.right, pr.Rect.bottom)
        End If
        If image0.Aoi.Count>0 Then
            image0.Aoi.GetBounds(rl,rt,rr,rb)
        End If

        Dim oldRange As Boolean = image0.UseActiveFrameRange
        If oldRange Then image0.UseActiveFrameRange = False

        'extract 1 frame
        image1 = image0.Duplicate(mcImageCreateFlags.mcicfNoAddToCollection Or mcImageCreateFlags.mcicfNotVisible)
        TmpImages.Add(image1)
        If oldRange Then image0.UseActiveFrameRange = oldRange

        Dim im As McImage = image1

        Dim nPhases As Integer=4
        Dim bs As McBusyState=ThisApplication.SetBusyState(,True)
        bs.SetProgressInfo("Measuring fiber thickness...", Interop.mcProgressAllowCancel.mcpacNestedCancelAllowed, , , nPhases)
        bs.BeginPhase()
        If opt.ImageType=eImageType.Native Then
            'binarize image
            binImage=BinarizeImage(image1)
            If bPreviewBin Then
                Dim ww As Object=ThisApplication.ActiveWindow
                'binImage.Visible=True
                'copy to clipboard
                binImage.Display.EditCopy(MediaCy.IQL.Display.Viewer.mcWhatCF_Types.mcwcftCF_DIB Or MediaCy.IQL.Display.Viewer.mcWhatCF_Types.mcwcftMcImageAoi)

                Dim p As MediaCy.IQL.ObjectManager.LONGPOINT
                If rl<>0 OrElse rt<>0 Then
                    p.x=rl'pr.Rect.left
                    p.y=rt'pr.Rect.top
                End If
                'paste to active image
                ThisApplication.ActiveWindow.View.Display.EditPasteAoi(,p,50)

                MsgBox("Check if the binarization is correct and click Ok.",vbOkOnly,"Binarization Preview")
                'paste stop
                ThisApplication.ActiveImage.ImageToolsOverlay.SelectTool()
                'binImage.Modified=False
                'binImage.Close
                ClearTempImages()
                're-activate
                ww.Activate
                Exit Sub
            End If
        Else
            'use it as Binary image
            binImage=image1
        End If
        bs.EndPhase()
        'work with binary image
        Dim ptsIm As McImage
        Dim rgns As McFeatures

        bs.BeginPhase()
        rgns=AllStepsOnBin(binImage)
        bs.EndPhase()

        Dim aoi As MediaCy.IQL.Features.McRegions = image0.Aoi
        If aoi.Count > 0 Then
            Dim r As MediaCy.IQL.ObjectManager.SINGLERECT = aoi.BoundingRect
            rgns.Move(r.left, r.top)
        End If

        bs.BeginPhase()
        SetFeatures(image0,rgns)
        bs.EndPhase()

        bs.BeginPhase()
        If opt.CollectToDC OrElse bFolder Then
            CollectDC()
        End If
        ClearTempImages()
        bs.EndPhase()

        'ShowHist
        bs.Dispose
    End Sub

    Public Function AllStepsOnBin(im As McImage) As McFeatures
        Dim ptsIm As McImage
        Dim dirArray As ArrayList=Nothing'array of vectors
        If Not opt.UseCircles Then
            dirArray = New ArrayList'array of vectors
        End If
        ptsIm=MeasureThicknessOnBin(im,dirArray)
        If opt.UseCircles Then
            Dim rgns As McRegions
            rgns=PointsToCircles(ptsIm)
            Return rgns
        Else
            Dim rgns As McLines
            rgns=PointsToLines(ptsIm,dirArray)
            Return rgns
        End If
    End Function

    Public Sub ShowHist
        With Measure.Measurements.Gadgets.Histogram(Nothing)
            .CheckState = MediaCy.IQL.Application.McCommand.mcCheckState.Checked
            .Run()
        End With
    End Sub

    Dim bShowImages As Boolean=False
    Private Function GetCreateFlags As mcImageCreateFlags
        Dim Flags As mcImageCreateFlags = mcImageCreateFlags.mcicfDefault
        If Not bShowImages Then
            Flags =  mcImageCreateFlags.mcicfNoAddToCollection Or mcImageCreateFlags.mcicfNotVisible
        End If
        Return Flags
    End Function

    Public Function MeasureThicknessOnBin(im As McImage,Optional ByRef dirArray As ArrayList=Nothing) As McImage
        Dim MeasureCentralThicknessOnBin2 As New SimpleScript
        Dim doc1, image1 As McImage, doc2, image2 As McImage, doc3, image3 As McImage, doc4
        Dim Flags As mcImageCreateFlags = GetCreateFlags
        Dim image0 As McImage
        image0=im
        image1=image0.Duplicate(Flags)
        TmpImages.Add(image1)

        image1.Filter3D.Thin3D(128)

        Dim MinBranchLength As Integer=opt.MinFiberLength

        image1.Filter3D.Prune3D(128,MinBranchLength,True)

        image1.Filter3D.BranchEnd3D(128,MediaCy.IQL.Filters.Volume.mcBranch3DMode.mcbm3DSkeleton)

        If opt.CentralLine Then
            image1=RemoveSegmentsOutsideCenter(image1)
        End If

        'shrink branches (offset from crossing)
        image1.Filter3D.Prune3D(1,MinBranchLength/4,False)

        Dim pts As McPoints
        'load options
        With Measure.Measurements.OptionsCommands.Open(MeasureCentralThicknessOnBin2)
            .FileName = ThisApplication.Path(mcPathType.mcptConfigurationFiles) & "FiberLines.iqo"
            .FilterIndex = 1
            .Run(image1)
        End With

        With Measure.MeasurementsCommands.ExecuteCount(MeasureCentralThicknessOnBin2)
            .Run(image1) 'now we have lines
        End With

        pts = LinesToPoints(image1, dirArray)

        'go to original binary image and create distance map
        image3 =image0.Filter3D.DistanceMap3D(128,1,1,1,Flags)
        TmpImages.Add(image3)

        SetPoints(image3,pts)

        Return image3
    End Function

    Private Function RemoveSegmentsOutsideCenter(im As McImage) As McImage
        'remove segments that do not cross central horizontal line
		With Measure.Measurements.OptionsCommands.Open(Nothing)
			.FileName = ThisApplication.Path(mcPathType.mcptConfigurationFiles) & "Central.iqo"
			.FilterIndex = 1
			.Run(im)
		End With

		With Measure.MeasurementsCommands.ExecuteCount(Nothing)
			.Run(im)
		End With

        Dim md As McMMData=im.MeasurementsData
        Dim pixsize As Double=1
        If im.SpatialCalibration IsNot Nothing Then
            pixsize=im.SpatialCalibration.PixelSizeY
        End If
        Dim ly As Double=im.Height/2'central line
        For Each sf As McMMSubFeature In md.SubFeatures
            Dim by2 As Double=0.5*sf.Value(eMeasures.RgnBoundingBoxHeight)/pixsize
            Dim cy As Double=sf.Value(eMeasures.RgnPixelCentroidY)/pixsize
            If (cy+by2)>ly And ((cy-by2)<ly) Then
                'segment crosses the line
            Else
                'segment doesn't cross the line, hide it
                sf.Show=False
            End If
        Next
        Dim im2 As McImage
        im2=md.CreateMask(mcImageCreateFlags.mcicfNotVisible Or mcImageCreateFlags.mcicfNoAddToCollection)
        im.Close
        Return im2
    End Function


    Public Function ObjectsToPoints(im As McImage) As McPoints
        If im Is Nothing Then Return Nothing
        Dim pixSize As Double=1
        If im.SpatialCalibration IsNot Nothing Then
            pixSize=im.SpatialCalibration.PixelSizeX
        End If
        Dim md As McMMData=im.MeasurementsData
        Dim rgns As MediaCy.IQL.Features.McPoints=ThisApplication.CreateOperator("McPoints",im)
        For Each sf As McMMSubFeature In md.SubFeatures
            Dim x As Double=sf.Value(eMeasures.PtLocationX)/pixSize
            Dim y As Double=sf.Value(eMeasures.PtLocationY)/pixSize
            Dim p As SINGLEPOINT
            p.x=x
            p.y=y
            rgns.SetFeaturePoints(rgns.Count,p)
        Next
        Return rgns
    End Function

    Public Function LinesToPoints(im As McImage,Optional ByRef dirArray As ArrayList=Nothing) As McPoints
        If im Is Nothing Then Return Nothing
        Dim md As McMMData=im.MeasurementsData
        Dim rgns As MediaCy.IQL.Features.McPoints=ThisApplication.CreateOperator("McPoints",im)
        Dim dStep As Double=opt.IncrStep
        Dim arr As System.Collections.ArrayList
        If opt.MeasureType=eMeasureType.OnePerSegment Then
            dStep=0'calc step automatically
        End If
        For Each sf As MediaCy.Addins.Measurements.McMMSubFeature In md.SubFeatures
            Dim drArray As ArrayList=Nothing'array of vectors per feature
            If dirArray IsNot Nothing Then
                drArray = New ArrayList'array of vectors
            End If
            arr =GenEvenDistancePoints(sf, dStep,drArray)
            Dim n As Integer=arr.Count
            Dim n1 As Integer=1
            Dim n2 As Integer=n-2'skip first and last
            If dStep=0 Then
                n1=n/2'just one point, in the middle
                n2=n1
            End If
            For i As Integer=n1 To n2
                rgns.SetFeaturePoints(rgns.Count,arr(i))
                If dirArray IsNot Nothing Then
                   dirArray.Add(drArray(i))
                End If
            Next
        Next
        Return rgns
    End Function

    Private Function CalcDirVector(p1 As SINGLEPOINT,p2 As SINGLEPOINT) As SINGLEPOINT
        Dim p As SINGLEPOINT
        p.y=p1.x-p2.x
        p.x=-(p1.y-p2.y)'rotate by 90 degrees
        Dim ln As Double=System.Math.Sqrt(p.x*p.x+p.y*p.y)
        If ln>0 Then
            p.x/=ln'normalize
            p.y/=ln
        Else
            p.x=1'fall back
            p.y=0
        End If
        Return p
    End Function


Function GenEvenDistancePoints(sf As McMMSubFeature, fStep As Double, Optional ByRef dirArray As ArrayList=Nothing) As ArrayList
    Dim distance As Single, delta As Single, dTrav As Single, dStep As Single
    Dim fractDist As Single, tmpF As Single
    Dim scale As Single
    Dim deltX As Single, deltY As Single
    Dim offX As Single, offY As Single
    Dim PtList() As SINGLEPOINT=Nothing
    Dim i As Integer, numPoints As Integer, Count As Integer
    Dim aoiType As Integer
    Dim nextPt As Integer, currPt As Integer, nextSamp As Integer

    Dim f As MediaCy.IQL.Features.McFeatures=sf.GetFeatures
    numPoints=f.GetFeaturePoints(sf.FeatureIndex,PtList)

    ' Calculate distance around the AOI
    distance = 0.0
    For i=0 To numPoints-2'open line
        nextPt = i+1
        If nextPt = numPoints Then
            nextPt = 0
        End If

        deltX = PtList(i).x - PtList(nextPt).x
        deltY = PtList(i).y - PtList(nextPt).y
        distance = distance + System.Math.Sqrt( deltX * deltX + deltY * deltY )
    Next i

    If fStep=0 Then
        If distance>150 Then
            delta = distance / 15
        Else
            delta = distance / 5
        End If
    Else
        delta = fStep'distance / numInterv
    End If

    ' Now walk around the list at intervals of delta,
    ' adding points as we go
    Dim outArray As New ArrayList
    'outArray.Add(PtList(0))
    nextSamp = 1

    currPt = 0
    nextPt = 1
    deltX = PtList(nextPt).x - PtList(currPt).x
    deltY = PtList(nextPt).y - PtList(currPt).y
    distance = System.Math.Sqrt( deltX * deltX + deltY * deltY )
    dStep = distance
    dTrav = 0.0

    ' This is an inchworm or bootstrap approach;
    ' move along the outline, adjusting one or
    ' the other criteria (placement or point list)
    ' until the proper number of points are set.
    Count = 0
    Dim p As SINGLEPOINT
    Do
        ' Two cases - we have traveled far enough for
        ' a new pair of points, or we haven't
        If (distance-dTrav) > 0.0 Then
            ' Add a new sample, update dTrav
            fractDist = (distance-dTrav) / dStep
            tmpF = PtList(nextPt).x - (deltX * fractDist)'+ 0.5
            p.x=tmpF
            tmpF = PtList(nextPt).y - (deltY * fractDist)' + 0.5
            p.y=tmpF
            outArray.Add(p)
            dTrav = dTrav + delta
            Count = Count + 1
        Else
            ' Shift to the next points, update total
            ' distance around AOI
            currPt = currPt + 1
            nextPt = nextPt + 1
            If nextPt = numPoints Then
                nextPt = 0
            End If

            deltX = PtList(nextPt).x - PtList(currPt).x
            deltY = PtList(nextPt).y - PtList(currPt).y
            dStep = System.Math.Sqrt( deltX * deltX + deltY * deltY )
            distance = distance + dStep
        End If

    Loop Until (currPt=numPoints) Or (nextPt=0)

    If dirArray IsNot Nothing Then
        Dim dr As SINGLEPOINT
        For ii As Integer=0 To outArray.Count-1
            If ii<1 Then
                dr=CalcDirVector(outArray(ii),PtList(0))'between the point and start
            Else If ii<outArray.Count-2
                dr=CalcDirVector(outArray(ii+1),outArray(ii-1))'one before, one after
            Else
                dr=CalcDirVector(outArray(ii),outArray(ii-1))'last point
            End If
            dirArray.Add(dr)
        Next
    End If

    Return outArray
End Function



    Private Function InRange(v As Double, r As Double, mx As Double) As Boolean
        Return ((v-r) > 0) AndAlso ((v+r) < mx)
    End Function

    Public Function PointsToCircles(im As McImage) As McRegions
        If im Is Nothing Then Return Nothing
        Dim pixSize As Double=1
        If im.SpatialCalibration IsNot Nothing Then
            pixSize=im.SpatialCalibration.PixelSizeX
        End If
        Dim md As McMMData=im.MeasurementsData
        Dim w As Integer=im.Width
        Dim h As Integer=im.Height
        Dim rgns As McRegions=ThisApplication.CreateOperator("McRegions",im)
        For Each sf As McMMSubFeature In md.SubFeatures
            Dim dens As Double=sf.Value(eMeasures.PtDensityUncalib)
            Dim d As Double=dens*2
            Dim x As Double=sf.Value(eMeasures.PtLocationX)/pixSize
            Dim y As Double=sf.Value(eMeasures.PtLocationY)/pixSize
            If InRange(x,dens,w) AndAlso InRange(y,dens,h) Then
                rgns.SetEllipse(rgns.Count,x,y,d,d,0)
            End If
        Next
        Return rgns
    End Function

    Private Sub SetLinePoints(ByRef pts() As SINGLEPOINT,x As Double, y As Double, r As Double, dr As SINGLEPOINT)
        pts(0).x=x-r*dr.x
        pts(0).y=y-r*dr.y
        pts(1).x=x+r*dr.x
        pts(1).y=y+r*dr.y
    End Sub


    Public Function PointsToLines(im As McImage,ByVal dirArray As ArrayList) As McLines
        If im Is Nothing Then Return Nothing
        Dim pixSize As Double=1
        If im.SpatialCalibration IsNot Nothing Then
            pixSize=im.SpatialCalibration.PixelSizeX
        End If
        Dim md As McMMData=im.MeasurementsData
        Dim w As Integer=im.Width
        Dim h As Integer=im.Height
        Dim rgns As McLines=ThisApplication.CreateOperator("McLines",im)
        Dim pts(1) As SINGLEPOINT
        Dim i As Integer=0
        For Each sf As McMMSubFeature In md.SubFeatures
            Dim dens As Double=sf.Value(eMeasures.PtDensityUncalib)
            Dim d As Double=dens*2
            Dim x As Double=sf.Value(eMeasures.PtLocationX)/pixSize
            Dim y As Double=sf.Value(eMeasures.PtLocationY)/pixSize
            If InRange(x,dens,w) AndAlso InRange(y,dens,h) Then
                If dirArray.Count>i Then
                    SetLinePoints(pts,x,y,d/2,dirArray(i))
                End If
                rgns.SetFeaturePoints(rgns.Count,pts)
                'rgns.SetEllipse(rgns.Count,x,y,d,d,0)
            End If
            i+=1
        Next
        Return rgns
    End Function


    Public Sub SetFeatures(im As McImage, rgns As McFeatures)
        If im Is Nothing OrElse rgns Is Nothing Then Exit Sub
        Dim nFeatures As Integer=rgns.Count
        With Measure.Measurements.OptionsCommands.Open(Nothing)
            .FileName = ThisApplication.Path(mcPathType.mcptConfigurationFiles) & "FiberThickness.iqo"
            .FilterIndex = 1
            .Run(im)
        End With
        If nFeatures>200 Then
            'switch off labels
            With Measure.MeasurementsCommands.Options(Nothing)
                .LabelType=MediaCy.Addins.Measurements.McMMOptions.mcmmltTypes.mcmmltNone
                .Run(im)
            End With
        End If

        Dim md As McMMData=im.MeasurementsData
        md.Reset
        Try
            'Dim rgns1 As MediaCy.IQL.Features.McRegions=ThisApplication.CreateOperator("McRegions",im)
            'rgns1.CopyFrom(rgns)'copy to new regions to avoid issues with rgns.AccessedImage when the image is closed
            md.AddFeaturesAsManualMeasurements(rgns)
            'rgns1.Reset
            rgns.Reset
        Catch Ex As System.Exception
            MsgBox(Ex.ToString)
        Finally
        End Try
    End Sub

    Public Sub SetPoints(im As McImage, rgns As McPoints)
        Dim md As McMMData=im.MeasurementsData
        md.Reset
        Try
            md.AddFeaturesAsManualMeasurements(rgns)
        Catch Ex As System.Exception
            MsgBox(Ex.ToString)
        Finally
        End Try
    End Sub

    Public Sub CalibrateImage()
        Dim CalibrateImage As New SimpleScript
        Dim doc1

        With Application.DocumentCommands.Active(CalibrateImage)
            .Run(doc1)
        End With

        With Measure.Calibration.SpatialCommands.QuickCalibration(CalibrateImage)
            .Interactive = True
            .Run(doc1)
        End With
        Dim im As McImage = ThisApplication.ActiveImage
        If im Is Nothing Then Exit Sub
        If im.SpatialCalibration Is Nothing Then
            opt.CalibrationName = ""
        Else
            opt.CalibrationName = im.SpatialCalibration.Name
        End If
    End Sub

    Public Function AddROI() As Boolean
        Dim im As McImage = ThisApplication.ActiveImage
        If im Is Nothing Then Return False
        Dim ROIs As McRegions = im.Aoi
        If ROIs.Count > 0 Then
            'prompt for name
            Dim defName As String = String.Format("Box AOI {0}", opt.ROIsList.ROIs.Count + 1)
            Dim Name As String = InputBox("Preset ROI name", "Name", defName)
            Name = Name.Replace(";", "_").Replace(",", "_") 'ensure that we don't have separators
            If Name.Length > 0 Then
                Dim r1 As Single,r2 As Single,r3 As Single,r4 As Single
                ROIs.GetFeatureBoundingBox(0, r1, r2, r3, r4)
                Dim r As SINGLERECT
                r.left=r1
                r.top=r2
                r.right=r3
                r.bottom=r4
                Dim pr As New PresetROI
                pr.SetRect(Name,r)
                opt.ROIsList.ROIs.Add(pr)
                opt.ActivePresetROI = opt.ROIsList.ROIs.Count
                Return True
            End If
        End If
        Return False
    End Function

    Public Function RemoveROI() As Boolean
        If opt.ActivePresetROI > 0 AndAlso (opt.ActivePresetROI - 1) < opt.ROIsList.ROIs.Count Then
            opt.ROIsList.ROIs.RemoveAt(opt.ActivePresetROI - 1)
            opt.ActivePresetROI = 0
            Return True
        End If
        Return False
    End Function
    Public Sub ActivateSelectionMode()
        Dim doc1

        With Mediacy.Automation.Application.DocumentCommands.Active(Nothing)
            .Run(doc1)
        End With

        With Measure.MeasurementsCommands.Unlock(Nothing)
            .Lock = False
            .Run(doc1)
        End With

        With Measure.Measurements.ToolsCommands.Select(Nothing)
            .Tool = eMMTool.SelectionTool
            .Run(doc1)
        End With
    End Sub

    Public Sub SelectFolder()
        Dim fd As New Microsoft.WindowsAPICodePack.Dialogs.CommonOpenFileDialog
        With fd
            If opt.SourceFolder <> "" Then .InitialDirectory = opt.SourceFolder
            .IsFolderPicker = True
            Dim result As System.Windows.Forms.DialogResult = .ShowDialog
            If result = System.Windows.Forms.DialogResult.OK Then
                opt.SourceFolder = .FileName
                opt.Save
            End If
        End With
    End Sub

    Public Sub SetupDC()
        Dim SetupDCm = New SimpleScript

        Measure.Data.CollectorCommands.Clear(SetupDCm).Run()

        With Measure.Data.Collector.MeasurementsCommands.Open(SetupDCm)
            .FileName = ThisApplication.Path(mcPathType.mcptConfigurationFiles) & "FiberMeasures.adoxsd"
            .Run()
        End With
        _DCisSet = True
    End Sub

    Public Sub ResetDC()
        SetupDC()
    End Sub

    Public Function CollectData() As SimpleScript
        CollectData = New SimpleScript
        Dim image1

        With Application.DocumentCommands.ActiveImage(CollectData)
            .Run(image1)
        End With

        With Measure.Data.CollectorCommands.Collect(CollectData)
            .Run(New List({image1}))
        End With

    End Function

    Private _DCisSet As Boolean = False
    Public Sub CollectDC()
        If Not _DCisSet Then
            SetupDC
        End If
        CollectData
    End Sub

    Public Function GetOutPath() As String
        Dim outPath
        ' User Code Here
        If opt.SourceFolder <> "" AndAlso opt.OutputFolder <> "" Then
            outPath = opt.SourceFolder & "\" & opt.OutputFolder & "\"
        Else
            outPath = ThisApplication.Path(mcPathType.mcptWriteableDocuments)
        End If
        If Not System.IO.Directory.Exists(outPath) Then
            System.IO.Directory.CreateDirectory(outPath)
        End If
        Return outPath
    End Function


    Public Function ExportDC() As SimpleScript
        ExportDC = New SimpleScript

        Dim outPath

        With Automate.ScriptingCommands.CodeCommand(ExportDC)
            If .Run() Then
                ' User Code Here
                outPath = GetOutPath
            End If
        End With

        With Measure.Data.Collector.Gadgets.DataTable(ExportDC)
            .CheckState = MediaCy.IQL.Application.McCommand.mcCheckState.Checked
            .Run()
        End With

        With Measure.Data.Collector.TableCommands.ActivateTab(ExportDC)
            .TabName = "MeasurementStats"
            .Run()
        End With

        With Measure.Data.Collector.TableCommands.SaveAsText(ExportDC)
            .FileName = outPath & "FiberStats.xls"
            .UseStatistics = False
            .Run()
        End With

    End Function

    Public Sub ShowDC()
        With Measure.Data.Collector.Gadgets.DataTable(Nothing)
            .CheckState = MediaCy.IQL.Application.McCommand.mcCheckState.Checked
            .Run()
        End With
    End Sub

    Public Sub BatchRunBefore()
        SetupDC
    End Sub

    Public Sub SaveActiveData()
        'save data to XLS file
        SaveMeasData
    End Sub

    Public Function SnapImage(inImage As McImage) As McImage
        Dim window1, image1, doc1
        'snap only active frame
        inImage.UseActiveFrameRange = False

        With Application.WindowCommands.Active(Nothing)
            .Run(window1)
        End With

        With Application.WindowCommands.Snap(Nothing)
            .Visible = False
            .Run(window1, image1)
        End With

        'restore
        inImage.UseActiveFrameRange = True
        Return image1
    End Function

    Private Function GetOutName(im As McImage, ext As String) As String
        Dim imName As String = im.File.FullPathName
        Dim fileName As String = System.IO.Path.Combine(GetOutPath, System.IO.Path.GetFileNameWithoutExtension(imName) + "." + ext)
        Return fileName
    End Function

    Public Sub SaveActiveSnapshot()
        Dim im As McImage = SnapImage(ThisApplication.ActiveImage)
        'save image
        Dim fn As String = GetOutName(ThisApplication.ActiveImage, "jpg")
        Try
            im.SaveAs(fn)
        Catch
        Finally
        End Try
    End Sub

    Public Sub BatchLoopOn()
        MeasureFiberThickness(True)
        If opt.SaveData Then
            SaveActiveData
        End If
        If opt.SaveSnapshot Then
            SaveActiveSnapshot
        End If
    End Sub

    Public Sub BatchRunAfter()
        ExportDC
    End Sub

    Public Sub RunBatch()
        If String.IsNullOrEmpty(opt.SourceFolder) Then
            MsgBox("Batch processing folder is not defined. Please define batch folder and try again.", vbExclamation Or vbOkOnly, "No Folder")
            Exit Sub
        End If

        With Automate.ScriptingCommands.BatchMacro(Nothing)
            .BeforeMacroFile = "FiberThickness>*Module1.vb*"
            .BeforeMacroName = "BatchRunBefore"
            .LoopMacroFile = "FiberThickness>*Module1.vb*"
            .LoopMacroName = "BatchLoopOn"
            .AfterMacroFile = "FiberThickness>*Module1.vb*"
            .AfterMacroName = "BatchRunAfter"
            .DocumentFormats = "*.tif;*.tiff;*.bmp;*.seq;*.jpg;*.jpe;*.jpeg;*.jp2;*.jpf;*.png;*.avi"
            .FolderToProcess = opt.SourceFolder
            .CloseDocuments = Not opt.KeepResultImage
            .DisplayDocuments = True
            .Run()
        End With
    End Sub

    Public Sub SaveMeasData()
        With Measure.Measurements.TableCommands.SaveAsText(Nothing)
            .FileName = GetOutName(ThisApplication.ActiveImage, "xls")
            .FilterIndex = 3
            .UseStatistics = False
            .Run(ThisApplication.ActiveDocument)
        End With
    End Sub

    Public Sub ShowLocalZoom
        With Application.Gadgets.LocalZoom(Nothing)
            .CheckState = MediaCy.IQL.Application.McCommand.mcCheckState.Checked
            .Run()
        End With
    End Sub

End Module
